﻿package  net.nicoptere.bitmap.vectorize
{
	import flash.display.Graphics;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import net.nicoptere.geometry.Spline;
	
	/**
	 * http://en.nicoptere.net/
	 * @author nicoptere
	 */
	public class VectorShape 
	{
		
		
		private var _color:uint;
		private var _rect:Rectangle;
		private var _points:Vector.<Point>;
		private var _simplified:Vector.<Point>;
		
		/**
		 * data holder to gather color, bounding rect, set of points and simplified version of the path
		 * @param	color the shape's color
		 * @param	rect the bounding rect
		 * @param	points the points of the shape's path 
		 * @param	simplified (optional) a simplified version of the path
		 */
		public function VectorShape( color:uint, rect:Rectangle, points:Vector.<Point>, simplified:Vector.<Point> = null ) 
		{
			
			_color = color;
			_rect = rect.clone();
			_points = points.concat();
			_simplified = ( simplified == null ) ? null : simplified.concat();
			
		}
		
		
		/**
		 * renders the shape to the given graphics
		 * @param	graphics the graphics to draw to
		 */
		public function draw( graphics:Graphics, smooth:Boolean = false ):void
		{
			
			var pts:Vector.<Point>;
			//tries to use the simplified path if no simplified version is available, it takes the normal one
			if ( _simplified != null )
			{
				if ( smooth ) 
				{
					pts = Spline.CubicPath(  _simplified, true );
				}else {
					pts = _simplified;
				}
				
			}
			else 
			{
				if ( smooth ) 
				{
					pts = Spline.CubicPath(  _points, true );
				}else {
					pts = _points;
				}
			}
			
			
			graphics.beginFill( color );
			graphics.moveTo( rect.x + pts[0].x, rect.y + pts[0].y );
			
			var i:int = -1;
			while ( i++ < pts.length-1 )
			{
				graphics.lineTo( rect.x + pts[i].x, rect.y + pts[i].y );
			}
			
			//graphics.lineTo( rect.x + pts[0].x, rect.y + pts[0].y );
			graphics.endFill();
			
		}
		
		
		
		//accessors
		
		/**
		 * assigns the shape's color
		 */
		public function set color( color:uint ):void 
		{
			_color = color;
		}
		/**
		 * retrieves the shape's color
		 */
		public function get color():uint { return _color; }
		
		
		
		/**
		 * assigns the shape's rect 
		 */
		public function set rect(rect:Rectangle):void 
		{
			_rect = rect.clone();
		}
		/**
		 * retrieves the shape's rect 
		 */
		public function get rect():Rectangle { return _rect; }
		
		
		
		/**
		 * assigns the shape's points
		 */
		public function set points( points:Vector.<Point> ):void 
		{
			_points = points.concat();
		}
		/**
		 * retrieves the shape's points 
		 */
		public function get points():Vector.<Point> { return _points; }
		
		
		
		/**
		 * assigns the shape's simplified set of points
		 */
		public function set simplified( points:Vector.<Point> ):void 
		{
			_simplified = points.concat();
		}
		/**
		 * retrieves the shape's simplified set of points 
		 */
		public function get simplified():Vector.<Point> { return _simplified; }
	}
	
}