﻿package  net.nicoptere.bitmap.vectorize
{
	/**
	 * http://en.nicoptere.net/
	 * @author nicoptere
	 */
	
	import flash.display.BitmapData;
	import flash.display.Graphics;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.utils.getTimer;
	
	import net.sakri.flash.bitmap.MarchingSquares
	import net.nicoptere.bitmap.vectorize.VectorShape; 
	
	public class Vectorize
	{
		
		//minimum area  for a shape to be valid
		public var minimumShapeArea:int = 4;
		
		
		//the BitmapData
		private var _src:BitmapData;
		
		
		//holds a list of shapes
		public var shapes:Vector.<VectorShape>;
		
		
			//optimization parameters to simplify the path
			//function by Andreas Weber
			//http://www.motiondraw.com/blog/?p=50
			
			//wether to optimize or not
			private var simplify:Boolean;
			
			//lookahead distance
			private var distance:int = 10;

			//tolerance
			private var tolerance:Number = 2;
			
			
			//infos
			private var pointCount:uint = 0;
			private var simpleCount:uint = 0;
			private var processTime:uint = 0;
		
		/**
		 * 
		 * @param	src the bitmapData to process
		 * @param	simplification an object containing the distance:Number & tolerance:Number values for simplification
		 */
		public function Vectorize( src:BitmapData = null, simplify:Boolean = true, simplification:Object = null ) 
		{
			
			if ( src != null )
			{
				
				bitmapData = src;
				
			}
			
			this.simplify = simplify;
			
			if ( simplification != null ) 
			{
				
				distance = !isNaN( simplification.distance ) ? 10 : simplification.distance;
				
				tolerance = !isNaN( simplification.distance ) ? 2 : simplification.tolerance;
				
			}
			
		}
		///accessors
		public function set bitmapData( src:BitmapData ):void 
		{
			_src = new BitmapData( src.width, src.height, true, 0x00000000 );
			_src.draw( src );
		}
		public function get bitmapData():BitmapData { return _src; }
		
		
		/**
		 * collects the eligibles shapes vectorization
		 * @param simplify a boolean to tell wether to simplify or not
		 * @param simplification an object containing the distance:Number & tolerance:Number values for simplification 
		 * @return an array of VectorShape objects
		 */
		public function vectorize( simplify:Boolean = false, simplification:Object = null ):Vector.<VectorShape>
		{
			
			if ( _src == null )
			{
				trace( 'no bitmapData to process' );
				return null;
			}
			
			this.simplify = simplify;
			
			if ( simplification != null ) 
			{
				
				distance = ( simplification.distance == null ) ? 10 : simplification.distance;
				
				tolerance = ( simplification.tolerance == null ) ? 2 : simplification.tolerance;
				
			}
			
			//flushes the previous shapes
			shapes = new Vector.<VectorShape>();
			
			return process( _src.clone() );
			
		}
		
		
		private function process( src:BitmapData ):Vector.<VectorShape>
		{
			
			//color of the component
			var ocol:uint;	
			
			//bounding box of the component
			var rect:Rectangle;
			
			//cordinates of the pathes
			var points:Vector.<Point> = new Vector.<Point>();
			
				//eventually creates an simplified version
				if ( simplify ) var simple:Vector.<Point> = new Vector.<Point>();
			
			
			//temp vars
			var col:uint = 0;	
			var pt:Point = new Point();
			var tmpBd:BitmapData = _src.clone();
			
			//stats purpose
			pointCount = 0;
			simpleCount = 0;
			var t:uint = getTimer();
			
			
			var W:int = src.width;		
			var length:int = src.width * src.height;
			
			var i:int;		
			var x:int = 0;
			var y:int = 0;
			
			
			//recursive loop through the image 
			src.lock();
			for ( i = 0; i < length; i++ )
			{
				
				x = ( i % W );
				y = int( i / W );
				
				//component'original color
				ocol = src.getPixel32( x , y );
				
				
				//forcing the white ... can't get what happens ...
				if ( ocol.toString( 16 ) == 'ffffffff' ) ocol = 0xFFFFFF; //<jeeze! why's that necessary?
				
				//if not transparent 
				if(  ocol != 0 )
				{
					
					//uses a temp random color to floodfill the pixel
					col = getColor();
					src.floodFill( i % W, int( i / W ) , col );
					
					//the 'component' is isolated via the colorbounds rect
					rect = src.getColorBoundsRect( 0xFFFFFFFF,  col );
					
					
					//if the rect matches the minimu size
					if ( ( rect.width * rect.height ) > minimumShapeArea )
					{
						//heavy 
						//creates a tmp bitmapData to isolate the component only
						tmpBd.fillRect( tmpBd.rect, 0x00000000 );
						tmpBd.copyPixels( src, rect, pt );
						
						//hmmmm isolates the shape in the tmpBd... doesn't do much
						//tmpBd.threshold( tmpBd, rect, pt, '!=', ocol, 0x00000000 );
						
						//performs the marching squares algorithm
						points = MarchingSquares.getBlobOutlinePointsClockwise( tmpBd );
						pointCount += points.length;
						
						
						//if we checked simplify
						if ( simplify )
						{
							simple = simplifyLang( distance, tolerance, points )
							simpleCount += simple.length;
						}
						
						//adds to the list
						shapes.push( new VectorShape( ocol, rect, points, simple ) );
						
					}
					//floodfills with transparency
					src.floodFill( x, y , 0x00000000 );
					
				}
			}
			processTime = ( getTimer() - t );
			return shapes;
			
		}
		
		
		/***********************************************
		 * 
		 * UTILS
		 * 
		***********************************************/
		
		/**
		 * renders the shape to the given graphics
		 * @param	graphics the graphics to draw to
		 */
		public function draw( graphics:Graphics, smooth:Boolean = false ):void 
		{
			var i:int = 0;
			var shape:VectorShape;
			for each ( shape in shapes )
			{
				shape.draw( graphics, smooth );
			}
			
		}
		
		
		//function by Andreas Weber to simplify the path
		//http://www.motiondraw.com/blog/?p=50
		static private function simplifyLang( distance:int, tolerance:Number, points:Vector.<Point> ):Vector.<Point>
		{
			
			if (distance <= 1)
			{
				return points;
			};
			var offset:Number, len:Number, count:Number;
			var PL:int = points.length;
			
			var i:int;
			var tmp:Vector.<Point> = new Vector.<Point>();
			tmp.push( points[0] );
			
			count = 1;
			
			if (distance > PL - 1)
			{
				distance = PL - 1;
			};
			
			
			for ( i = 0; i < PL; i++)
			{
				if (i + distance > PL) 
				{ 
					distance = PL - i -1 
				};
				
				offset = recursiveToleranceBar(points, i, distance, tolerance);
				if (offset > 0)
				{
					tmp.push( points[i + offset] );
					i += offset-1;
					count++;
				}
			}
			tmp.push( points[0] );
			return tmp;
		}
		
		
		// this function is called by simplifyLang
		static private function recursiveToleranceBar(points:Vector.<Point>, c:int, distance:int, tolerance:Number):int
		{
			var n:int = distance;
			var cp:Point = points[c];
			var cLP:Point = new Point();
			
			var v1:Point = new Point();
			var v2:Point = new Point(); 
			var angle:Number, dx:Number, dy:Number;
			
			var id:int 
			if ( c + n > points.length - 1 ) 
			{
				id = points.length - 1;
				n =  points.length - 1 - c;
			}else {
				id = ( c + n );
			}
			
			v1 = new Point( points[ id ].x - cp.x, points[id].y - cp.y );
			var dist:Number;
			var j:int;
			
			for ( j = 1; j <= n; j++)
			{
				cLP = points[c+j]; 
				  
				v2.x = cLP.x - cp.x;
				v2.y = cLP.y - cp.y;
				  
				angle = Math.acos((v1.x * v2.x + v1.y * v2.y)/(Math.sqrt(v1.y * v1.y + v1.x * v1.x)*Math.sqrt(v2.y * v2.y + v2.x * v2.x)));
				 
				 
				if(isNaN(angle)){angle = 0;}
				
				dx = cp.x - cLP.x; 
				dy = cp.y - cLP.y;
				
				dist = Math.sqrt(dx*dx+dy*dy);

				if ( Math.sin(angle) * dist >= tolerance)
				{
					n--;
					if (n > 0)
					{	
						return recursiveToleranceBar(points, c, n, tolerance);
					}else{
						return 0;
					}
				}
			}
			return n;
		}
		
		
		/**
		 * returns a random() color for components labelling
		 * @return a random uint 
		 */
		private function getColor():uint
		{
			return 0xFF << 24 | Math.random() * 0xFFFFFF;
		}
		
		public function get info():String 
		{ 
			return 'it took ' + processTime + ' ms to process ' + shapes.length + ' shapes (' + pointCount + ' points / ' + simpleCount + ' when simplified )'; 
		}
	}
}
